from sqlalchemy import Column, Text, DateTime, text, Integer, Index, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property

Base = declarative_base()


class GDCReport(Base):
    __tablename__ = "gdc_reports"
    __table_args__ = (
        Index("{}_report_idx".format(__tablename__), "report", postgresql_using="gin"),
        Index("{}_report_type_idx".format(__tablename__), "report_type"),
        Index("{}_created_datetime_idx".format(__tablename__), "created_datetime"),
        Index("{}_program_idx".format(__tablename__), "program"),
        Index("{}_project_idx".format(__tablename__), "project"),
        Index("{}_id_idx".format(__tablename__), "id"),
    )

    def __repr__(self):
        return "<Report({}, {})>".format(self.id, self.report_type)

    id = Column(Integer, primary_key=True)
    program = Column(Text)
    project = Column(Text)
    report = Column(JSONB)
    report_type = Column(Text, nullable=False)

    created_datetime = Column(
        DateTime(timezone=True),
        nullable=False,
        server_default=text("now()"),
    )

    @hybrid_property
    def project_id(self):
        return self.program + "-" + self.project

    @project_id.expression
    def project_id(cls):
        return func.concat(cls.program, "-", cls.project)



from dictionaryutils import dictionary as gdcdictionary
import psqlgraph
import sqlalchemy


class GDCGraphValidator(object):
    """
    Validator that validates entities' relationship with existing nodes in
    database.

    """

    def __init__(self):
        self.schemas = gdcdictionary
        self.required_validators = {"links_validator": GDCLinksValidator()}
        self.optional_validators = {}

    def record_errors(self, graph, entities):
        for validator in self.required_validators.values():
            validator.validate(entities, graph)

        for entity in entities:
            schema = self.schemas.schema[entity.node.label]
            validators = schema.get("validators")
            if validators:
                for validator_name in validators:
                    self.optional_validators[validator_name].validate()


class GDCLinksValidator(object):
    def validate(self, entities, graph=None):
        for entity in entities:
            for link in gdcdictionary.schema[entity.node.label]["links"]:
                if "name" in link:
                    self.validate_edge(link, entity)
                elif "subgroup" in link:
                    self.validate_edge_group(link, entity)

    def validate_edge_group(self, schema, entity):
        submitted_links = []
        schema_links = []
        num_of_edges = 0

        for group in schema["subgroup"]:
            if "subgroup" in group:
                # nested subgroup
                result = self.validate_edge_group(group, entity)
            elif "name" in group:
                result = self.validate_edge(group, entity)
            else:
                raise Exception(
                    f"This link subgroup should include one of 'name' or 'subgroup', but both are missing: {group}"
                )

            if result["length"] > 0:
                submitted_links.append(result)
                num_of_edges += result["length"]
            schema_links.append(result["name"])

        if schema.get("required") is True and len(submitted_links) == 0:
            names = ", ".join(schema_links[:-2] + [" or ".join(schema_links[-2:])])
            entity.record_error(
                "Entity is missing a required link to {}".format(names),
                keys=schema_links,
            )

        if schema.get("exclusive") is True and len(submitted_links) > 1:
            names = ", ".join(schema_links[:-2] + [" and ".join(schema_links[-2:])])
            entity.record_error(
                "Links to {} are exclusive.  More than one was provided.".format(
                    schema_links
                ),
                keys=schema_links,
            )

        result = {"length": num_of_edges, "name": ", ".join(schema_links)}
        return result

    def validate_edge(self, link_sub_schema, entity):
        association = link_sub_schema["name"]
        node = entity.node
        targets = node[association]
        result = {"length": len(targets), "name": association}

        if len(targets) > 0:
            multi = link_sub_schema["multiplicity"]

            if multi in ["many_to_one", "one_to_one"]:
                if len(targets) > 1:
                    entity.record_error(
                        "'{}' link has to be {}".format(association, multi),
                        keys=[association],
                    )

            if multi in ["one_to_many", "one_to_one"]:
                for target in targets:
                    if len(target[link_sub_schema["backref"]]) > 1:
                        entity.record_error(
                            "'{}' link has to be {}, target node {} already has {}".format(
                                association,
                                multi,
                                target.label,
                                link_sub_schema["backref"],
                            ),
                            keys=[association],
                        )

            if multi == "many_to_many":
                pass
        else:
            if link_sub_schema.get("required") is True:
                entity.record_error(
                    "Entity is missing required link to {}".format(association),
                    keys=[association],
                )
        return result


class GDCUniqueKeysValidator(object):
    def validate(self, entities, graph=None):
        for entity in entities:
            schema = gdcdictionary.schema[entity.node.label]
            node = entity.node
            for keys in schema["uniqueKeys"]:
                props = {}
                if keys == ["id"]:
                    continue
                for key in keys:
                    prop = schema["properties"][key].get("systemAlias")
                    if prop:
                        props[prop] = node[prop]
                    else:
                        props[key] = node[key]
                if graph.nodes(type(node)).props(props).count() > 1:
                    entity.record_error(
                        "{} with {} already exists in the GDC".format(
                            node.label, props
                        ),
                        keys=props.keys(),
                    )



# -*- coding: utf-8 -*-
"""
gen3datamodel.gdc_postgres_admin
----------------------------------

Module for stateful management of a GDC PostgreSQL installation.
"""

import argparse
import logging
import random
import sqlalchemy as sa
import time

from collections import namedtuple
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

#: Required but 'unused' import to register GDC models
from . import models  # noqa

from psqlgraph import (
    create_all,
    Node,
    Edge,
)

logging.basicConfig()
logger = logging.getLogger("gdc_postgres_admin")
logger.setLevel(logging.INFO)

name_root = "table_creator_"
app_name = "{}{}".format(name_root, random.randint(1000, 9999))

GRANT_READ_PRIVS_SQL = """
BEGIN;
GRANT SELECT ON TABLE {table} TO {user};
COMMIT;
"""

GRANT_WRITE_PRIVS_SQL = """
BEGIN;
GRANT SELECT, INSERT, UPDATE, DELETE ON TABLE {table} TO {user};
COMMIT;
"""

REVOKE_READ_PRIVS_SQL = """
BEGIN;
REVOKE SELECT, INSERT, UPDATE, DELETE ON TABLE {table} FROM {user};
COMMIT;
"""

REVOKE_WRITE_PRIVS_SQL = """
BEGIN;
REVOKE INSERT, UPDATE, DELETE ON TABLE {table} FROM {user};
COMMIT;
"""


def execute(engine, sql, *args, **kwargs):
    statement = sa.sql.text(sql)
    logger.debug(statement)
    return engine.execute(statement, *args, **kwargs)


def get_engine(host, user, password, database):
    connect_args = {"application_name": app_name}
    con_str = "postgresql://{user}:{pwd}@{host}/{db}".format(
        user=user, host=host, pwd=password, db=database
    )
    return create_engine(con_str, connect_args=connect_args)


def execute_for_all_graph_tables(engine, sql, *args, **kwargs):
    """Execute a SQL statment that has a python format variable {table}
    to be replaced with the tablename for all Node and Edge tables

    """
    for cls in Node.__subclasses__() + Edge.__subclasses__():
        _kwargs = dict(kwargs, **{"table": cls.__tablename__})
        statement = sql.format(**_kwargs)
        execute(engine, statement)


def grant_read_permissions_to_graph(engine, user):
    execute_for_all_graph_tables(engine, GRANT_READ_PRIVS_SQL, user=user)


def grant_write_permissions_to_graph(engine, user):
    execute_for_all_graph_tables(engine, GRANT_WRITE_PRIVS_SQL, user=user)


def revoke_read_permissions_to_graph(engine, user):
    execute_for_all_graph_tables(engine, REVOKE_READ_PRIVS_SQL, user=user)


def revoke_write_permissions_to_graph(engine, user):
    execute_for_all_graph_tables(engine, REVOKE_WRITE_PRIVS_SQL, user=user)


def migrate_transaction_snapshots(engine, user):
    """
    Updates to TransactionSnapshot table:
        - change old `id` column to `entity_id`, which is no longer unique or primary
          key
        - add new serial `id` column as primary key
    """
    md = MetaData(bind=engine)
    tablename = models.submission.TransactionSnapshot.__tablename__
    snapshots_table = Table(tablename, md, autoload=True)
    if "entity_id" not in snapshots_table.c:
        execute(
            engine,
            "ALTER TABLE {name} DROP CONSTRAINT {name}_pkey".format(name=tablename),
        )
        execute(
            engine,
            "ALTER TABLE {} RENAME id TO entity_id".format(tablename),
        )
        execute(
            engine,
            "ALTER TABLE {} ADD COLUMN id SERIAL PRIMARY KEY;".format(tablename),
        )


def create_graph_tables(engine, timeout):
    """
    create a table
    """
    logger.info("Creating tables (timeout: %d)", timeout)

    connection = engine.connect()
    trans = connection.begin()
    logger.info("Setting lock_timeout to %d", timeout)

    timeout_str = "{}s".format(int(timeout + 1))
    connection.execute("SET LOCAL lock_timeout = %s;", timeout_str)

    create_all(connection)
    trans.commit()


def create_tables(engine, delay, retries):
    """Create the tables but do not kill any blocking processes.

    This command will catch OperationalErrors signalling timeouts from
    the database when the lock was not obtained successfully within
    the `delay` period.

    """

    logger.info("Running table creator named %s", app_name)
    try:
        return create_graph_tables(engine, delay)

    except OperationalError as e:
        if "timeout" in str(e):
            logger.warning("Attempt timed out")
        else:
            raise

        if retries <= 0:
            raise RuntimeError("Max retries exceeded")

        logger.info(
            "Trying again in {} seconds ({} retries remaining)".format(delay, retries)
        )
        time.sleep(delay)

        create_tables(engine, delay, retries - 1)


def subcommand_create(args):
    """Idempotently/safely create ALL tables in database that are required
    for the GDC.  This command will not delete/drop any data.

    """

    logger.info("Running subcommand 'create'")
    engine = get_engine(args.host, args.user, args.password, args.database)
    kwargs = dict(
        engine=engine,
        delay=args.delay,
        retries=args.retries,
    )

    return create_tables(**kwargs)


def subcommand_grant(args):
    """Grant permissions to a user.

    Argument ``--read`` will grant users read permissions
    Argument ``--write`` will grant users write and READ permissions
    """

    logger.info("Running subcommand 'grant'")
    engine = get_engine(args.host, args.user, args.password, args.database)

    assert args.read or args.write, "No premission types/users specified."

    if args.read:
        users_read = [u for u in args.read.split(",") if u]
        for user in users_read:
            grant_read_permissions_to_graph(engine, user)

    if args.write:
        users_write = [u for u in args.write.split(",") if u]
        for user in users_write:
            grant_write_permissions_to_graph(engine, user)


def subcommand_revoke(args):
    """Grant permissions to a user.

    Argument ``--read`` will revoke users' read permissions
    Argument ``--write`` will revoke users' write AND READ permissions
    """

    logger.info("Running subcommand 'revoke'")
    engine = get_engine(args.host, args.user, args.password, args.database)

    if args.read:
        users_read = [u for u in args.read.split(",") if u]
        for user in users_read:
            revoke_read_permissions_to_graph(engine, user)

    if args.write:
        users_write = [u for u in args.write.split(",") if u]
        for user in users_write:
            revoke_write_permissions_to_graph(engine, user)


def add_base_args(subparser):
    subparser.add_argument(
        "-H", "--host", type=str, action="store", required=True, help="psql-server host"
    )
    subparser.add_argument(
        "-U", "--user", type=str, action="store", required=True, help="psql test user"
    )
    subparser.add_argument(
        "-D",
        "--database",
        type=str,
        action="store",
        required=True,
        help="psql test database",
    )
    subparser.add_argument(
        "-P",
        "--password",
        type=str,
        action="store",
        default="",
        help="psql test password",
    )
    return subparser


def add_subcommand_create(subparsers):
    parser = add_base_args(
        subparsers.add_parser("graph-create", help=subcommand_create.__doc__)
    )
    parser.add_argument(
        "--delay",
        type=int,
        action="store",
        default=60,
        help="How many seconds to wait for blocking processes to finish before retrying.",
    )
    parser.add_argument(
        "--retries",
        type=int,
        action="store",
        default=10,
        help="If blocked by important process, how many times to retry after waiting `delay` seconds.",
    )


def add_subcommand_grant(subparsers):
    parser = add_base_args(
        subparsers.add_parser("graph-grant", help=subcommand_grant.__doc__)
    )
    parser.add_argument(
        "--read",
        type=str,
        action="store",
        help="Users to grant read access to (comma separated).",
    )
    parser.add_argument(
        "--write",
        type=str,
        action="store",
        help="Users to grant read/write access to (comma separated).",
    )


def add_subcommand_revoke(subparsers):
    parser = add_base_args(
        subparsers.add_parser("graph-revoke", help=subcommand_revoke.__doc__)
    )
    parser.add_argument(
        "--read",
        type=str,
        action="store",
        help="Users to revoke read access from (comma separated).",
    )
    parser.add_argument(
        "--write",
        type=str,
        action="store",
        help=(
            "Users to revoke write access from (comma separated). "
            "NOTE: The user will still have read privs!!"
        ),
    )


def get_parser():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subcommand")
    add_subcommand_create(subparsers)
    add_subcommand_grant(subparsers)
    add_subcommand_revoke(subparsers)
    return parser


def main(args=None):
    args = args or get_parser().parse_args()

    logger.info("[ HOST     : %-10s ]", args.host)
    logger.info("[ DATABASE : %-10s ]", args.database)
    logger.info("[ USER     : %-10s ]", args.user)

    return_value = {
        "graph-create": subcommand_create,
        "graph-grant": subcommand_grant,
        "graph-revoke": subcommand_revoke,
    }[args.subcommand](args)

    logger.info("Done.")
    return return_value



# -*- coding: utf-8 -*-

"""gen3datamodel.models
----------------------------------

This module defines all of the ORM classes for the GDC datamodel
python layer.  Classes are produced by a factory, and required
attributes are injected into them before they are registered into the
modules globals map.

::WARNING:: This code is the heart of the GDC.  Changes here will
propogate to all code that imports this package and MAY BREAK THINGS.

- jsm

"""
from cdislogging import get_logger
from collections import defaultdict
from dictionaryutils import dictionary
from .misc import FileReport  # noqa
from sqlalchemy.orm import configure_mappers
from .versioned_nodes import VersionedNode  # noqa

import hashlib
from . import versioned_nodes  # noqa
from . import notifications
from . import submission

from sqlalchemy import func, event, and_
from psqlgraph import Node, Edge, pg_property

from sqlalchemy.ext.hybrid import (
    Comparator,
    hybrid_property,
)

from .indexes import (
    cls_add_indexes,
    get_secondary_key_indexes,
)

logger = get_logger("gen3datamodel")

# Deprecated; used only for logging a deprecation notice
CACHE_CASES = (
    True
    if (not hasattr(dictionary, "settings") or not dictionary.settings)
    else dictionary.settings.get("enable_case_cache", True)
)
if CACHE_CASES:
    logger.info("Caching related cases is deprecated")

# These are properties that are defined outside of the JSONB column in
# the database, inform later code to skip these
excluded_props = ["id", "type"]


# At module load time, evaluate which classes have already been
# registered as subclasses of the abstract bases Node and Edge to
# prevent double-registering
loaded_nodes = [c.__name__ for c in Node.get_subclasses()]
loaded_edges = [c.__name__ for c in Edge.get_subclasses()]


def remove_spaces(s):
    """Returns a stripped string with all of the spaces removed.

    :param str s: String to remove spaces from

    """
    return s.replace(" ", "")


def register_class(cls):
    """Register a class in `globals`.  This allows us to import the ORM
    classes from :mod:`gen3datamodel.models`

    :param cls: Any class object
    :returns: None

    """

    globals()[cls.__name__] = cls


def get_links(schema):
    """Given a schema, pull out all of the ``links`` that this type can
    have an edge to.

    :returns: a ``dict`` of format ``{<name>: <link>}``

    """
    result = {}

    def _recursive_get_links(links, result):
        for entry in links:
            if "subgroup" in entry:
                _recursive_get_links(entry["subgroup"], result)
            else:
                result[entry["name"]] = entry

    _recursive_get_links(schema.get("links", []), result)
    return result


def types_from_str(types):
    return [
        a
        for type_ in types
        for a in {
            "string": [str],
            "number": [float, int],
            "integer": [int],
            "float": [float],
            "null": [type(None)],
            "boolean": [bool],
            "array": [list],
            None: [str],
        }[type_]
    ]


def PropertyFactory(name, schema, key=None):
    """Returns a pg_property (psqlgraph specific type of hybrid_property)"""
    key = name if key is None else key

    # Assert the dictionary has no references for properties
    assert "$ref" not in schema.keys(), (
        "Found a JSON reference in dictionary.  These should be resolved "
        "at gdcdictionary module load time as of 2016-02-24"
    )

    # Lookup property type and coerce to list
    types = schema.get("type")
    types = [types] if not isinstance(types, list) else types

    # Convert the list of string type identifiers to Python types
    python_types = types_from_str(types)

    # If there is an enum defined, grab it for pg_property validation
    enum = schema.get("enum")

    # Create pg_property setter
    @pg_property(*python_types, enum=enum)
    def setter(self, val):
        self._set_property(key, val)

    setter.__name__ = name

    return setter


def get_class_name_from_id(_id):
    return "".join([a.capitalize() for a in _id.split("_")])


def get_class_tablename_from_id(_id):
    return "node_{}".format(_id.replace("_", ""))


def cls_inject_versioned_nodes_lookup(cls):
    """Injects a property and a method into the class to retrieve node
    versions.

    """

    @property
    def _versions(self):
        """Returns a query if the node is bound to a session. Raises an
          exception if the node is not bound to a session.

        :returns: A SQLAlchemy query for node versions.

        """

        session = self.get_session()
        if not session:
            raise RuntimeError(
                "{} not bound to a session. Try .get_versions(session).".format(self)
            )
        return self.get_versions(session)

    def get_versions(self, session):
        """Returns a query for node versions given a session."""

        return (
            session.query(VersionedNode)
            .filter(VersionedNode.node_id == self.node_id)
            .filter(VersionedNode.label == self.label)
            .order_by(VersionedNode.key.desc())
        )

    cls._versions = _versions
    cls.get_versions = get_versions


def cls_inject_created_datetime_hook(
    cls, updated_key="updated_datetime", created_key="created_datetime"
):
    """Given a class, inject a SQLAlchemy hook that will write the
    timestamp of the last session flush to the :param:`updated_key`
    and :param:`created_key` properties.

    """

    @event.listens_for(cls, "before_insert")
    def set_created_updated_datetimes(mapper, connection, target):
        ts = target.get_session()._flush_timestamp.isoformat("T")
        if updated_key in target.props:
            target._props[updated_key] = ts
        if created_key in target.props:
            target._props[created_key] = ts


def cls_inject_updated_datetime_hook(cls, updated_key="updated_datetime"):
    """Given a class, inject a SQLAlchemy hook that will write the
    timestamp of the last session flush to the :param:`updated_key`
    property.

    """

    @event.listens_for(cls, "before_update")
    def set_updated_datetimes(mapper, connection, target):
        # check if any of the attributes have changed before updating the node
        if target.get_session().is_modified(target, include_collections=False):
            ts = target.get_session()._flush_timestamp.isoformat("T")
            if "updated_datetime" in target.props:
                target._props["updated_datetime"] = ts


def cls_inject_secondary_keys(cls, schema):
    """The dictionary defines a list of ``unique`` keys.  If there are
    keys (possibly tuples of keys) in addition to the canonical `id`
    column, then we want to be able to query the database against
    these keys.

    This function injects:
        0. a list of ``str`` keys that should be unique to
           ``_pg_secondary_keys``
        0. ``_secondary_keys_dict`` which is a dictionary of str key
           to :class:`sqlalchemy.dialects.postgresql.json.JSONElement`
           values
        0. ``_secondary_keys``, a tuple of
           :class:`sqlalchemy.dialects.postgresql.json.JSONElement`
           objects

    """

    unique_keys = schema.get("uniqueKeys", [])
    cls.__pg_secondary_keys = [keys for keys in unique_keys if "id" not in keys]

    class SecondaryKeyComparator(Comparator):
        def __eq__(self, other):
            filters = []
            cls = self.__clause_element__()
            secondary_keys = getattr(cls, "__pg_secondary_keys", [])
            for keys, values in zip(secondary_keys, other):
                if "id" in keys:
                    continue
                for key, val in zip(keys, values):
                    filters.append(
                        func.lower(cls._props[key].astext) == func.lower(val)
                    )
            return and_(*filters)

    @property
    def _secondary_keys_dicts(self):
        vals = []
        secondary_keys = getattr(self, "__pg_secondary_keys", [])
        for keys in secondary_keys:
            if "id" in keys:
                continue
            vals.append({key: getattr(self, key, None) for key in keys})
        return vals

    @hybrid_property
    def _secondary_keys(self):
        vals = []
        for keys in getattr(self, "__pg_secondary_keys", []):
            vals.append(tuple(getattr(self, key) for key in keys))
        return tuple(vals)

    @_secondary_keys.comparator
    def _secondary_keys(cls):
        return SecondaryKeyComparator(cls)

    # Set this attribute so psqlgraph doesn't treat it as a property
    _secondary_keys._is_pg_property = False
    cls._secondary_keys = _secondary_keys
    cls._secondary_keys_dicts = _secondary_keys_dicts

    cls_add_indexes(cls, get_secondary_key_indexes(cls))


def NodeFactory(_id, schema):
    """Returns a node class given a schema."""

    name = get_class_name_from_id(_id)
    links = get_links(schema)

    @property
    def node_id(self, value):
        return self.node_id

    @node_id.setter
    def node_id(self, value):
        self.node_id = value

    # Pull the JSONB properties from the `properties` key
    attributes = {
        key: PropertyFactory(key, schema)
        for key, schema in schema.get("properties", {}).items()
        if key not in links and key not in excluded_props
    }

    # Store for the programmer
    attributes["_dictionary"] = {
        "category": schema.get("category"),
        "title": schema.get("title"),
    }

    # _pg_links are out_edges, links TO other types
    attributes["_pg_links"] = {}

    # _pg_backrefs are in_edges, links FROM other types
    attributes["_pg_backrefs"] = {}

    # _pg_edges are all edges, links to AND from other types
    attributes["_pg_edges"] = {}

    # Create the Node subclass!
    cls = type(
        name,
        (Node,),
        dict(
            __tablename__=get_class_tablename_from_id(_id),
            __label__=_id,
            id=node_id,
            **attributes
        ),
    )

    cls_inject_created_datetime_hook(cls)
    cls_inject_updated_datetime_hook(cls)
    cls_inject_versioned_nodes_lookup(cls)
    cls_inject_secondary_keys(cls, schema)

    return cls


def generate_edge_tablename(src_label, label, dst_label):
    """Generate a name for the edge table.

    Because of the limit on table name length on PostgreSQL, we have
    to truncate some of the longer names.  To do this we concatenate
    the first 2 characters of each word in each of the input arguments
    up to 10 characters (per argument).  However, this strategy would
    very likely lead to collisions in naming.  Therefore, we take the
    first 8 characters of a hash of the full, un-truncated name
    *before* we truncate and prepend this to the truncation.  This
    gets us a name like ``edge_721d393f_LaLeSeqDaFrLaLeSeBu``.  This
    is rather an undesirable workaround. - jsm

    """

    tablename = "edge_{}{}{}".format(
        src_label.replace("_", ""),
        label.replace("_", ""),
        dst_label.replace("_", ""),
    )

    # If the name is too long, prepend it with the first 8 hex of it's hash
    # truncate the each part of the name
    if len(tablename) > 40:
        oldname = tablename
        logger.debug("Edge tablename {} too long, shortening".format(oldname))
        tablename = "edge_{}_{}".format(
            str(
                hashlib.md5(
                    tablename.encode("utf-8"), usedforsecurity=False
                ).hexdigest()
            )[:8],
            "{}{}{}".format(
                "".join([a[:2] for a in src_label.split("_")])[:10],
                "".join([a[:2] for a in label.split("_")])[:7],
                "".join([a[:2] for a in dst_label.split("_")])[:10],
            ),
        )
        logger.debug("Shortening {} -> {}".format(oldname, tablename))

    return tablename


def EdgeFactory(
    name,
    label,
    src_label,
    dst_label,
    src_dst_assoc,
    dst_src_assoc,
    _assigned_association_proxies=defaultdict(set),
):
    """Returns an edge class.

    :param name: The name of the edge class.
    :param label: Assigned to ``edge.label``
    :param src_label: The label of the source edge
    :param dst_label: The label of the destination edge
    :param src_dst_assoc:
        The link name i.e. ``src.src_dst_assoc`` returns a list of
        destination type nodes
    :param dst_src_assoc:
        The backref name i.e. ``dst.dst_src_assoc`` returns a list of
        source type nodes
    :param _assigned_association_proxies:
        Don't pass this parameter. This will be used to store what
        links and backrefs have been assigned to the source and
        destination nodes.  This prevents clobbering a backref with a
        link or a link with a backref, as they would be from different
        nodes, should be different relationships, and would have
        different semantic meanings.

    """

    # Correctly format all of the names
    name = remove_spaces(name)
    label = remove_spaces(label)
    src_label = remove_spaces(src_label)
    dst_label = remove_spaces(dst_label)
    src_dst_assoc = remove_spaces(src_dst_assoc)
    dst_src_assoc = remove_spaces(dst_src_assoc)

    # Generate the tablename. If it is too long, it will be hashed and
    # truncated.
    tablename = generate_edge_tablename(src_label, label, dst_label)

    # Lookup the tablenames for the source and destination classes
    src_cls = Node.get_subclass(src_label)
    dst_cls = Node.get_subclass(dst_label)

    # Assert that we're not clobbering link names
    assert dst_src_assoc not in _assigned_association_proxies[dst_label], (
        "Attempted to assign backref '{link}' to node '{node}' but "
        "the node already has an attribute called '{link}'".format(
            link=dst_src_assoc, node=dst_label
        )
    )
    assert src_dst_assoc not in _assigned_association_proxies[src_label], (
        "Attempted to assign link '{link}' to node '{node}' but "
        "the node already has an attribute called '{link}'".format(
            link=src_dst_assoc, node=src_label
        )
    )

    # Remember that we're adding this link and this backref
    _assigned_association_proxies[dst_label].add(dst_src_assoc)
    _assigned_association_proxies[src_label].add(src_dst_assoc)

    hooks_before_insert = Edge._session_hooks_before_insert

    hooks_before_update = Edge._session_hooks_before_update

    hooks_before_delete = Edge._session_hooks_before_delete

    cls = type(
        name,
        (Edge,),
        {
            "__label__": label,
            "__tablename__": tablename,
            "__src_class__": get_class_name_from_id(src_label),
            "__dst_class__": get_class_name_from_id(dst_label),
            "__src_dst_assoc__": src_dst_assoc,
            "__dst_src_assoc__": dst_src_assoc,
            "__src_table__": src_cls.__tablename__,
            "__dst_table__": dst_cls.__tablename__,
            "_session_hooks_before_insert": hooks_before_insert,
            "_session_hooks_before_update": hooks_before_update,
            "_session_hooks_before_delete": hooks_before_delete,
        },
    )

    return cls


def load_nodes():
    """Parse all nodes from dictionary and create Node subclasses"""

    for entity, subschema in dictionary.schema.items():
        name = subschema["title"]
        _id = subschema["id"]
        if name not in loaded_nodes:
            try:
                cls = NodeFactory(_id, subschema)
            except Exception:
                print("Unable to load {}".format(name))
                raise
            else:
                register_class(cls)


def parse_edge(src_label, name, edge_label, subschema, link):
    """Parse an edge from the dictionary and create and Edge subclass

    :returns: The outbound name of the edge

    """

    dst_label = link["target_type"]
    backref = link["backref"]

    src_label = subschema["id"]
    if dst_label not in dictionary.schema:
        raise RuntimeError(
            "Destination '{}' for edge '{}' from '{}' not defined".format(
                dst_label, name, src_label
            )
        )

    dst_label = dictionary.schema[dst_label]["id"]
    edge_name = "".join(map(get_class_name_from_id, [src_label, edge_label, dst_label]))

    edge = EdgeFactory(
        edge_name,
        edge_label,
        src_label,
        dst_label,
        name,
        backref,
    )

    register_class(edge)

    return "_{}_out".format(edge_name)


def load_edges():
    """Add a dictionry of links from this class

    { <link name>: {'backref': <backref name>, 'type': <source type> } }

    """

    for src_label, subschema in dictionary.schema.items():

        src_cls = Node.get_subclass(src_label)
        if not src_cls:
            raise RuntimeError("No source class labeled {}".format(src_label))

        for name, link in get_links(subschema).items():
            edge_label = link["label"]
            edge_name = parse_edge(src_label, name, edge_label, subschema, link)
            src_cls._pg_links[link["name"]] = {
                "edge_out": edge_name,
                "dst_type": Node.get_subclass(link["target_type"]),
            }


def inject_pg_backrefs():
    """Add a dict of links to this class.  Backrefs look like:

    .. code-block::
        { <link name>: {'name': <backref name>, 'src_type': <source type> } }

    """

    for src_label, subschema in dictionary.schema.items():
        for name, link in get_links(subschema).items():
            dst_cls = Node.get_subclass(link["target_type"])
            dst_cls._pg_backrefs[link["backref"]] = {
                "name": link["name"],
                "src_type": Node.get_subclass(src_label),
            }


def inject_pg_edges():
    """Add a dict of ALL the links, to and from, each class

    .. code-block::
        { <link name>: {'backref': <backref name>, 'type': <target type> } }

    """

    def find_backref(link, src_cls):
        """Given the JSON link definition and a source class :param:`src_cls`,
        return the name of the backref

        """

        for prop, backref in link["dst_type"]._pg_backrefs.items():
            if backref["src_type"] == cls:
                return prop

    def cls_inject_forward_edges(cls):
        """We should have already added the links that go OUT from this class,
        so let's add them to `_pg_edges`

        :returns: None, cls is mutated

        """

        for name, link in cls._pg_links.items():
            cls._pg_edges[name] = {
                "backref": find_backref(link, cls),
                "type": link["dst_type"],
            }

    def cls_inject_backward_edges(cls):
        """We should have already added the links that go INTO this class,
        so let's add them to `_pg_edges`

        :returns: None, cls is mutated

        """

        for name, backref in cls._pg_backrefs.items():
            cls._pg_edges[name] = {
                "backref": backref["name"],
                "type": backref["src_type"],
            }

    for cls in Node.get_subclasses():
        cls_inject_forward_edges(cls)
        cls_inject_backward_edges(cls)


load_nodes()
load_edges()
inject_pg_backrefs()
inject_pg_edges()
configure_mappers()



import argparse
import getpass
import psqlgraph

from .models import *  # noqa
from psqlgraph import *  # noqa
from sqlalchemy import *  # noqa
from .models.versioned_nodes import VersionedNode  # noqa

try:
    import IPython

    ipython = True
except Exception as e:
    print(
        (
            "{}, using standard interactive console. "
            "If you install IPython, then it will automatically "
            "be used for this repl."
        ).format(e)
    )
    import code

    ipython = False


message = """
Entering psqlgraph console:
    database : {}
    host     : {}
    user     : {}

NOTE:
    PsqlGraphDriver stored in local variable `g`.
    A `g.session_scope` session is at `s`.
    `rb()` will rollback the session.
"""


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--database",
        default="test",
        type=str,
        help="name of the database to connect to",
    )
    parser.add_argument(
        "-i",
        "--host",
        default="localhost",
        type=str,
        help="host of the postgres server",
    )
    parser.add_argument(
        "-u", "--user", default="test", type=str, help="user to connect to postgres as"
    )
    parser.add_argument(
        "-p",
        "--password",
        default=None,
        type=str,
        help="password for given user. If no " "password given, one will be prompted.",
    )

    args = parser.parse_args()

    print(message.format(args.database, args.host, args.user))
    if args.password is None:
        args.password = getpass.getpass()

    g = psqlgraph.PsqlGraphDriver(args.host, args.user, args.password, args.database)

    with g.session_scope() as s:
        rb = s.rollback
        if ipython:
            IPython.embed()
        else:
            code.InteractiveConsole(locals=globals()).interact()
